/*
 * debug.c
 *
 * The debug routines
 *
 * Kekoa Proudfoot 1/2/97
 */

/*****************************************************************************
 * Include files
 *****************************************************************************/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "debug.h"

/*****************************************************************************
 * Macro definitions
 *****************************************************************************/

#define DEBUG_MAX_STRING_LENGTH  64

/*****************************************************************************
 * Global variables
 *****************************************************************************/

static char  libname[DEBUG_MAX_STRING_LENGTH] = DEBUG_DEFAULT_LIBNAME;
static int   debug_mode                       = DEBUG_DEFAULT_MODE;
/*static FILE *debug_file                       = DEBUG_DEFAULT_FILE;*/
static FILE *debug_file;

/*****************************************************************************
 * Interface function declarations
 *****************************************************************************/

void
debug_set_lib(char *lib)
{
    strncpy(libname, lib, DEBUG_MAX_STRING_LENGTH);
}

void
debug_set_mode(int mode)
{
    debug_mode = mode;
}

int
debug_get_mode(void)
{
    return debug_mode;
}

void
debug_set_file(FILE *file)
{
    debug_file = file;
}

FILE *
debug_get_file(void)
{
    return debug_file;
}

int
debug_check_mode(int mask)
{
    return ((debug_mode & mask) && debug_file);
}

void
debug_print(int mask, char *fmt, ...)
{
    va_list arglist;

    if (debug_check_mode(mask)) {
	va_start(arglist, fmt);

	if (libname && libname[0])
	    fprintf(debug_file, "%s: ", libname);

	vfprintf(debug_file, fmt, arglist);

	putc('\n', debug_file);

	va_end(arglist);
    }
}

void
debug_fatal(char *fmt, ...)
{
    va_list arglist;

    if (debug_file) {
	va_start(arglist, fmt);

	if (libname && libname[0])
	    fprintf(debug_file, "%s: ", libname);

	vfprintf(debug_file, fmt, arglist);

	putc('\n', debug_file);

	va_end(arglist);
    }

    exit(1);
}
