/*
 * debug.h
 *
 * Header file for the debug routines
 *
 * Kekoa Proudfoot 1/2/97
 */

#ifndef DEBUG_H_DEFINED
#define DEBUG_H_DEFINED

/*****************************************************************************
 * Include files
 *****************************************************************************/

#include <stdio.h>

/*****************************************************************************
 * Type definitions
 *****************************************************************************/

typedef enum debug_bit_t {
    DEBUG_NONE                   = 0x0000,
    DEBUG_WARNING                = 0x0001,
    DEBUG_PROXY_ADDRESS          = 0x0002,
    DEBUG_CONNECT_STATE          = 0x0004,
    DEBUG_CONTROL_PACKET_RECIPT  = 0x0008,
    DEBUG_GAME_PACKET_RECIPT     = 0x0010,
    DEBUG_SEQUENCE_NUMBERS       = 0x0020,
    DEBUG_INCOMING_PACKET_DATA   = 0x0040,
    DEBUG_OUTGOING_PACKET_DATA   = 0x0080,
    DEBUG_INVALID_PACKET_DATA    = 0x0100,
    DEBUG_INCOMING_TCP_DATA      = 0x0200,
    DEBUG_OUTGOING_TCP_DATA      = 0x0400,
    DEBUG_INCOMING_COMMANDS      = 0x0800,
    DEBUG_OUTGOING_COMMANDS      = 0x1000
} debug_bit_t;

/*****************************************************************************
 * Configurable macros
 *****************************************************************************/

#define DEBUG_DEFAULT_LIBNAME    "debug"
#define DEBUG_DEFAULT_MODE       ( DEBUG_WARNING               | \
				   DEBUG_PROXY_ADDRESS         | \
				   DEBUG_CONNECT_STATE         | \
				   DEBUG_CONTROL_PACKET_RECIPT | \
				   DEBUG_INVALID_PACKET_DATA )
#define DEBUG_DEFAULT_FILE       stderr

/*****************************************************************************
 * Interface function declarations
 *****************************************************************************/

extern void  debug_set_lib    (char *lib);
extern void  debug_set_mode   (int mode);
extern int   debug_get_mode   (void);
extern void  debug_set_file   (FILE *file);
extern FILE *debug_get_file   (void);
extern int   debug_check_mode (int mask);
extern void  debug_print      (int mask, char *fmt, ...);
extern void  debug_fatal      (char *fmt, ...);

#endif
