#include "main.h"

void 
process_server_packet(void)
{
    int length, position;

    if( !net_recv(gs.server_socket, &gs.recv_packet) ) {
      print_to_log("Recv error from quake server: %s\n", strerror(errno));
      send_to_clients("exception Recv error from quake server: %s\n", 
                       strerror(errno));
      reset_connection();
      return;
    }

    /* For PACKET_UNRELIABLE, PACKET_RELIABLE_MORE, and PACKET_RELIABLE_END,
     * check to make sure that the just received packet number is greater than
     * the packet number of the last received packet of that type. This is 
     * because UDP can duplicate packets, and we don't want to process them
     * twice.
     */
    switch( gs.recv_packet.type ) {
      case PACKET_CONTROL:
          if( debug_check_mode(DEBUG_INCOMING_PACKET_DATA) ) {
            packet_sv_to_cl_debug(&gs.recv_packet, gs.logfp);
          }
          process_control_packet();
          break;
      case PACKET_RELIABLE_REPLY:
          if( debug_check_mode(DEBUG_INCOMING_PACKET_DATA) ) {
            packet_sv_to_cl_debug(&gs.recv_packet, gs.logfp);
          }
          if( !queue_empty(&gs.send_reliable_queue) ) {
              packet_destroy((packet_t *)queue_remove(&gs.send_reliable_queue));
              send_reliable_head();
          }
          break;
      case PACKET_UNRELIABLE:
          if( debug_check_mode(DEBUG_INCOMING_PACKET_DATA) ) {
            packet_sv_to_cl_debug(&gs.recv_packet, gs.logfp);
          }
          if( gs.recv_packet.unreliable.number == 0 ||
              gs.recv_packet.unreliable.number > gs.recv_unreliable_number ) {
            length = 1;
            position = 0;
            while( length && position < gs.recv_packet.unreliable.length) {
                length = process_mesg(&gs.recv_packet.unreliable.data[position]);
                position += length;
            }
            send_unreliable_head();
            gs.recv_unreliable_number = gs.recv_packet.unreliable.number;
          }
          break;
      case PACKET_RELIABLE_MORE:
          if( gs.recv_packet.reliable.number == 0 ||
              gs.recv_packet.reliable.number >  gs.recv_reliable_number ) {
            memcpy((char *)&gs.recv_buffer.reliable.data[gs.recv_buffer_pos],
                    gs.recv_packet.reliable.data, gs.recv_packet.reliable.length);
            gs.recv_buffer_pos     += gs.recv_packet.reliable.length;
            gs.recv_reliable_number = gs.recv_packet.reliable.number;
          }
          if( debug_check_mode(DEBUG_INCOMING_PACKET_DATA) ) {
             packet_sv_to_cl_debug(&gs.recv_packet, gs.logfp);
          }
          send_acknowledge();
          break;
      case PACKET_RELIABLE_END:
          gs.post_signon = FALSE;
          if( gs.recv_packet.reliable.number == 0 ||
              gs.recv_packet.reliable.number >  gs.recv_reliable_number ) {
            memcpy((char *)&gs.recv_buffer.reliable.data[gs.recv_buffer_pos],
                    gs.recv_packet.reliable.data, gs.recv_packet.reliable.length);
            gs.recv_buffer_pos             += gs.recv_packet.reliable.length;
            gs.recv_buffer.reliable.length  = gs.recv_buffer_pos;
            gs.recv_buffer.reliable.type    = gs.recv_packet.reliable.type;
            gs.recv_buffer.reliable.number  = gs.recv_packet.reliable.number;
            if( debug_check_mode(DEBUG_INCOMING_PACKET_DATA) ) {
              packet_sv_to_cl_debug(&gs.recv_buffer, gs.logfp);
            }
            length = 1;
            position = 0;
            while( length && position < gs.recv_buffer.reliable.length) {
                length    = process_mesg(&gs.recv_buffer.reliable.data[position]);
                position += length;
            }
            gs.recv_buffer_pos      = 0;
            gs.recv_reliable_number = gs.recv_packet.reliable.number;
          }
          send_acknowledge();
          if( gs.post_signon ) {
            signon_protocol();
            gs.post_signon = FALSE;
          }
          break;
      default:
      case PACKET_UNDEFINED:
          fprintf(gs.logfp, "[server packet: %s, %d bytes]\n",
                  packet_name(&gs.recv_packet), gs.recv_packet.header.length);
          hex_dump((unsigned char *)&gs.recv_packet, gs.recv_packet.header.length, "S", gs.logfp);
          break;
    }
}

void 
process_control_packet(void)
{
    sv_to_cl_ctrl_t  ctrl;

    sv_to_cl_ctrl_get(gs.recv_packet.control.data, &ctrl);

    switch( ctrl.op ) {
       case CTRL_REP_ACCEPT:
          gs.server_personal_port = (unsigned int)ctrl.rep_accept.port;
          net_connect_socket(gs.server_socket, &gs.server_addr, (short)gs.server_personal_port);
          print_to_log("Quake connection accepted: %s\n", 
                                   net_address_to_name(&gs.server_addr));
          send_to_clients("commandprint Quake connection accepted: %s\n", 
                            net_address_to_name(&gs.server_addr));
          send_to_clients("server %s\n",     gs.server_name);
          break;
       case CTRL_REP_REJECT:
          print_to_log("Rejected: %s\n", map_weird_chars(ctrl.rep_reject.reason));
          send_to_clients("rejected Rejected: %s\n", map_weird_chars(ctrl.rep_reject.reason));
          break;
       default:
          break;
    }
}

int 
process_mesg(char *buf)
{
    int  len, i;
    sv_to_cl_mesg_t mesg;
 
    len = sv_to_cl_mesg_get(buf, &mesg);

    switch( mesg.type ) {
      case MESG_NOP:
          send_to_clients("print Server to Client ----> KeepAlive\n");
          print_to_log("Server to Client ----> KeepAlive\n");
          if( gs.keepalive_count++ > 4 ) {
            send_to_clients("print Retry connection after too many KeepAlives\n");
            print_to_log("Retry connection after too many KeepAlives\n");
            send_disconnect();
          }
          break;
      case MESG_DISCONNECT:
          send_to_clients("commandprint Received Disconnect from Server\n");
          print_to_log("Received Disconnect from Server\n");
          reset_connection();
          break;
      case MESG_TIME:
          gs.time = mesg.time.time;
          break;
      case MESG_PRINT:
          if( mesg.print.text[strlen(mesg.print.text)-1] == '\n' ) {
            strcat(gs.print_buf, map_weird_chars(mesg.print.text));
            gs.print_buf[strlen(gs.print_buf)-1] = '\n';
            parse_mesg_prints(gs.print_buf);
            gs.print_buf[0] = '\0';
          } else {
            if( strstr(mesg.print.text, "VERSION") &&
                strstr(mesg.print.text, "SERVER (")  &&
                strstr(mesg.print.text, "CRC)") ) {
              sprintf(gs.print_buf, "%s\n", 
                 map_weird_chars(mesg.print.text));
              parse_mesg_prints(gs.print_buf);
              gs.print_buf[0] = '\0';
            } else {
              strcat(gs.print_buf, map_weird_chars(mesg.print.text));
            }
          }
          break;
      case MESG_SERVER_INFO:
          gs.version     = mesg.server_info.version;
          gs.max_players = mesg.server_info.max_clients;
          strcpy(gs.mapname, map_weird_chars(mesg.server_info.map_name));
          send_to_clients("version %d\n", gs.version);
          send_to_clients("maxplayers %d\n", gs.max_players);
          send_to_clients("mapname %s\n", gs.mapname);
          print_to_stats_log("date             %s\n", gs.timestr);
          print_to_stats_log("mapname          %s\n", gs.mapname);
          break;
      case MESG_UPDATE_NAME:
          i = (int)mesg.update_name.player;
          if( i >= 0 && i < MAX_PLAYERS ) {
            raw_map_name(gs.players[i].raw_mapped_name, mesg.update_name.name);
            strcpy(gs.players[i].real_name, mesg.update_name.name);
            strcpy(gs.players[i].name, map_weird_chars(mesg.update_name.name));
            send_to_clients("playername %d %s\n", i, gs.players[i].name);
            send_to_clients("playername_raw %d %s\n", i, gs.players[i].raw_mapped_name);
            if( gs.players[i].valid ) {
              print_to_stats_log("player_name   %2d %-15s %s\n", i+1, 
                      gs.players[i].ipaddr, gs.players[i].name);
              if( gs.players[i].just_entered && strlen(gs.players[i].name) != 0 ) {
                gs.players[i].just_entered = FALSE;
                send_to_clients("print %d %s %s\n", i+1, 
                      gs.players[i].ipaddr, gs.players[i].name);
              }
            }
          }
          break;
      case MESG_UPDATE_FRAGS:
          i = (int)mesg.update_frags.player;
          if( i >= 0 && i < MAX_PLAYERS ) {
            gs.players[i].frags = mesg.update_frags.frags;
            send_to_clients("playerfrags %d %d\n", i, gs.players[i].frags);
            print_to_stats_log("player_frags  %2d %d\n", i+1, gs.players[i].frags);
          }
          break;
      case MESG_UPDATE_COLORS:
          i = (int)mesg.update_colors.player;
          if( i >= 0 && i < MAX_PLAYERS ) {
            gs.players[i].shirt = (int)mesg.update_colors.shirt_color;
            gs.players[i].pants = (int)mesg.update_colors.pants_color;
            send_to_clients("playershirt %d %d\n", i, gs.players[i].shirt);
            send_to_clients("playerpants %d %d\n", i, gs.players[i].pants);
          }
          break;
      case MESG_SET_PAUSE:
          if( mesg.set_pause.pause ) {
             send_to_clients("print Game Paused\n");
          } else {
             send_to_clients("print Game Unpaused\n");
          }
          break;
      case MESG_SIGNON_NUM:
          gs.signon_level = (int)mesg.signon_num.signon_num;
          gs.post_signon = TRUE;
          break;
      case MESG_CENTER_PRINT:
/*          print_to_log("Center Print: %s\n", mesg.center_print.text);*/
          break;
      case MESG_STUFF_TEXT:
          if( mesg.stuff_text.text[strlen(mesg.stuff_text.text)-1] == '\n' ) {
            strcat(gs.stuff_buf, map_nl_to_semi(mesg.stuff_text.text));
            gs.stuff_buf[strlen(gs.stuff_buf)-1] = '\n';
            add_stuff_command(gs.stuff_buf);
            gs.stuff_buf[0] = '\0';
          } else {
            strcat(gs.stuff_buf, map_nl_to_semi(mesg.stuff_text.text));
          }
          break;
      case MESG_INTERMISSION:
          break;
      case MESG_SET_VIEW:
          if( mesg.set_view.entity >  0 && 
              mesg.set_view.entity <= MAX_PLAYERS &&
              mesg.set_view.entity != gs.player_number )
          {
            gs.player_number = mesg.set_view.entity;
            send_to_clients("mynumber %d\n", gs.player_number);
            print_to_log("My player number is %d\n", gs.player_number);
            print_to_stats_log("qsmack_num    %2d\n", gs.player_number);
          }
          break;
      case MESG_FINALE:
      case MESG_UPDATE_STAT:
      case MESG_SOUND:
      case MESG_SET_ANGLE:
      case MESG_LIGHT_STYLE:
      case MESG_CLIENT_DATA:
      case MESG_STOP_SOUND:
      case MESG_VERSION:
      case MESG_PARTICLE:
      case MESG_DAMAGE:
      case MESG_SPAWN_STATIC:
      case MESG_SPAWN_BINARY:
      case MESG_SPAWN_BASELINE:
      case MESG_TEMP_ENTITY:
      case MESG_KILLED_MONSTER:
      case MESG_FOUND_SECRET:
      case MESG_SPAWN_STATIC_SOUND:
      case MESG_CD_TRACK:
      case MESG_SELL_SCREEN:
      case MESG_UPDATE_ENTITY:
         break;
      default:
         print_to_log("Bad server mesg received\n");
         sv_to_cl_mesg_debug(&mesg, gs.logfp);
         break;
    }

    return len;
}

void 
signon_protocol(void)
{
  switch( gs.signon_level ) {
    case 1:
       empty_send_queues();
       send_mesg_client_connect();
       send_console_command("prespawn");
       break;
    case 2:
       send_mesg_signon_2();
       break;
    case 3:
       send_console_command("begin");
       gs.keepalive_count = 0;
       gs.connect_state   = IN_THE_GAME;
       break;
    default:
       print_to_log("Found illegal signon state: %d\n", gs.signon_level);
       send_to_clients("exception Found illegal signon state: %d\n", gs.signon_level);
       break;
   }
}

void 
parse_mesg_prints(char *text)
{
   static int active = 0, flag, index;
   char qtimeStr[16];
   int i, hours, minutes, seconds;

   if( strncmp(text, "host:    ", 9) == 0 ||
       strncmp(text, "version: ", 9) == 0 ||
       strncmp(text, "tcp/ip:  ", 9) == 0 ||
       strncmp(text, "ipx:     ", 9) == 0 ||
       strncmp(text, "map:     ", 9) == 0 )
   {
     return;
   }

   if( sscanf(text, "players: %d active (%*d max)", &active) == 1 ) {
     flag = 0;
     for(i=0; i<MAX_PLAYERS; i++) {
       gs.players[i].valid = FALSE;
     }
     return;
   }

   if( active  ) {
     if( flag == 0 ) {
       flag = 1;
       if( sscanf(text, "#%2d%*17c%*d%s", &index, qtimeStr) != 2 ) {
         print_to_log("Couldn't convert status: %s", text);
         print_to_log("Server version could be too old\n");
       }
       gs.players[index-1].valid = 1;
       if( sscanf(qtimeStr, "%d:%d:%d", &hours, &minutes, &seconds) != 3 ) {
         print_to_log("Couldn't convert time: %s", qtimeStr);
         hours = minutes = seconds = 0;
       }
       gs.players[index-1].seconds = (hours * 3600) + (minutes * 60) + seconds;
       strcpy(gs.players[index-1].time, qtimeStr);
     } else {
       flag = 0;
       if( sscanf(strtok(text, ":"), "%s", gs.players[index-1].ipaddr) != 1 ) {
         print_to_log("Couldn't convert status: %s", text);
       }
       if( active == 1 ) {
         print_delta_status();
         if( is_in_the_game() && ban_ok() ) {
           ban_check();
         }
       }
       active--;
     }
   } else if( strncmp(text, "#qsmack_command# ", 17) == 0 ) {
     send_to_clients("print %s", text);
     print_to_stats_log("server_mesg      %s", text);
     process_special_admin_commands(&text[17]);
   } else {
     if( !check_for_say_print(text) ) {
       send_to_clients("print %s", text);
       print_to_stats_log("server_mesg      %s", text);
     }
   }
}

void
print_delta_status(void)
{
  int i;

  for(i=0; i<MAX_PLAYERS; i++) {
    if( gs.players_old[i].valid == FALSE && gs.players[i].valid == TRUE ) 
    {
      gs.players[i].just_entered = TRUE;
      print_to_stats_log("player_enter  %2d %-15s %s\n", i+1, 
                gs.players[i].ipaddr, gs.players[i].name);
      send_to_clients("playerip %d %s\n", i, gs.players[i].ipaddr);
    } else if( gs.players_old[i].valid == TRUE && gs.players[i].valid == FALSE ) 
    {
      print_to_stats_log("player_exit   %2d %s\n", i+1, gs.players_old[i].time);
      send_to_clients("playerexit %d\n", i);
    } else if( gs.players_old[i].valid == TRUE &&
               gs.players[i].valid == TRUE &&
               strcmp(gs.players_old[i].ipaddr, gs.players[i].ipaddr) != 0 )
    {
      gs.players[i].just_entered = TRUE;
      print_to_stats_log("player_exit   %2d %s\n", i+1, gs.players_old[i].time);
      print_to_stats_log("player_enter  %2d %-15s %s\n", i+1, 
               gs.players[i].ipaddr, gs.players[i].name);
      send_to_clients("playerexit %d\n", i);
      send_to_clients("playerip %d %s\n", i, gs.players[i].ipaddr);
    }
    gs.players_old[i] = gs.players[i];
  }
}

int
check_for_say_print(char *text)
{
  char tmp[50], name[MAX_NAMELEN+1], *mesg;
  int i, say, num;

  say = FALSE;
  num = 0;

  for(i=0; i<MAX_CLIENTS; i++) {
    if( gs.client[i].valid ) {
      sprintf(tmp, " %s:", gs.client[i].stagename);
      if( strncmp(tmp, text, strlen(gs.client[i].stagename)+2) == 0 ) {
        say = TRUE;
        num = gs.player_number;
        strcpy(name, gs.client[i].stagename);
        break;
      }
    }
  }

  if( say == FALSE ) {
    sprintf(tmp, " %s:", gs.player_name);
    if( strncmp(tmp, text, strlen(gs.player_name)+2) == 0 ) {
      say = TRUE;
      num = gs.player_number;
      strcpy(name, gs.player_name);
    }
  }

  if( say == FALSE ) {
    for(i=0; i<MAX_PLAYERS; i++) {
      if( gs.players[i].valid && strlen(gs.players[i].name) > 0 ) {
        sprintf(tmp, " %s:", gs.players[i].name);
        if( strncmp(tmp, text, strlen(gs.players[i].name)+2) == 0 ) {
          say = TRUE;
          num = i+1;
          strcpy(name, gs.players[i].name);
          break;
        }
      }
    }
  }

  if( say == TRUE ) {
    mesg = &text[strlen(name)+1];
    send_to_clients("sayprint #%-2d %-15s%s", num, name, mesg);
    print_to_stats_log("player_say    %2d %-15s%s", num, name, mesg);
  }

  return say;
}

void 
process_special_admin_commands(char *cmd)
{
  char regex[100];
  char banner[16];
  int  i, shirt, pants;

  cmd[strlen(cmd)-1] = '\0';

  if( strlen(cmd) == 0 ) {
    print_to_log("Special command is zero length, ignored.\n");
    send_to_clients("commandprint Special command is zero length, ignored.\n");
    return;
  }

  banner[0] = '\0';
  get_banner_from_cmd(cmd, banner);

  if( gs.allow_in_game_command == FALSE ) {
    send_console_command("tell \"%s\" QSmack Says: No Soup For You!", cmd);
    send_console_command("tell \"%s\" In game banning has been turned off!", cmd);
    return;
  }

  if( strcmp(cmd, "reread_banfile") == 0 ) {
    destroy_ban_lists();
    read_ban_file();
  } else if( map_ip_to_regex(cmd, regex) ) {
    add_ban_ip_item(cmd, banner, (char *)NULL);
    write_ban_file();
  } else if( strncmp(cmd, "ban_color", 9) == 0 ) {
    if( sscanf(cmd, "ban_color %d %d", &shirt, &pants) == 2 ) {
       if( shirt >=0 && shirt <= 13 && pants >= 0 && pants <= 13 ) {
         add_ban_color_item(shirt, pants);
         write_ban_file();
       } else {
         print_to_log("Bad colors in special command: %s\n", cmd);
       }
    } else {
      print_to_log("Bad colors in special command: %s\n", cmd);
    }
  } else if( strncmp(cmd, "ban", 3) == 0 ) {
    if( sscanf(cmd, "ban %d", &i) == 1 ) {
       i--;
       if( i >= 0 && i < MAX_PLAYERS && gs.players[i].valid ) {
         add_ban_ip_item(gs.players[i].ipaddr, banner, gs.players[i].name);
         write_ban_file();
       } else {
        i++;
        print_to_log("Bad player number in ban: %d\n", i);
       }
    } else {
      say_ban_list();
    }
  } else if( sscanf(cmd, "unban %d", &i) ) {
     remove_ban_item_by_num(i);
     write_ban_file();
  } else {
    print_to_log("Unknown special command: %s\n", cmd);
    send_to_clients("commandprint Unknown special command: %s\n", cmd);
  }
}

void 
get_banner_from_cmd(char *cmd, char *banner)
{
   int i, count;

   for(i=0, count=0; i<MAX_PLAYERS; i++) {
     if( strcmp(cmd, gs.players[i].name) == 0 && gs.players[i].valid ) {
       count++;
       strcpy(banner, gs.players[i].ipaddr);
     }
   }

   if( count == 0 ) {
     print_to_log("No players with admin command name: %s\n", cmd);
   } else if( count > 1 ) {
     print_to_log("More than one player with admin command name: %s\n", cmd);
     for(i=0; i<MAX_PLAYERS; i++) {
       if( strcmp(cmd, gs.players[i].name) == 0 && gs.players[i].valid ) {
         print_to_log("%s\n", gs.players[i].ipaddr);
       }
     }
   }
}

char *
map_nl_to_semi(unsigned char *str)
{
  static unsigned char tmp[MAX_STRING_LEN];
  unsigned char *s = tmp;
  static const u_char QuakeBet[] = 
      "*    *    ;* >**"  "[]0123456789****"
      " !\"#$%&'()*+,-./" "0123456789:;<=>?"
      "@ABCDEFGHIJKLMNO"  "PQRSTUVWXYZ[\\]^_"
      "`abcdefghijklmno"  "pqrstuvwxyz{|}~<"
      "     *     * >**"  "[]0123456789*   "
      " !\"#$%&'()*+,-./" "0123456789:;<=>?"
      "@ABCDEFGHIJKLMNO"  "PQRSTUVWXYZ[\\]^_"
      "`abcdefghijklmno"  "pqrstuvwxyz{|}~<";
  
  while( *str ) {
    *s++ = QuakeBet[*str++];
  }
  *s = '\0';

  return tmp;
}

char *
map_weird_chars(unsigned char *str)
{
  static unsigned char tmp[MAX_STRING_LEN];
  unsigned char *s = tmp;
  static const u_char QuakeBet[] = 
      "*    *     * >**"  "[]0123456789****"
      " !\"#$%&'()*+,-./" "0123456789:;<=>?"
      "@ABCDEFGHIJKLMNO"  "PQRSTUVWXYZ[\\]^_"
      "`abcdefghijklmno"  "pqrstuvwxyz{|}~<"
      "     *     * >**"  "[]0123456789*   "
      " !\"#$%&'()*+,-./" "0123456789:;<=>?"
      "@ABCDEFGHIJKLMNO"  "PQRSTUVWXYZ[\\]^_"
      "`abcdefghijklmno"  "pqrstuvwxyz{|}~<";
  
  while( *str ) {
    *s++ = QuakeBet[*str++];
  }
  *s = '\0';

  return tmp;
}

void
raw_map_name(unsigned char *dst, unsigned char *src)
{
   unsigned char tmp[5];
   unsigned int i;

   dst[0] = '\0';

   for(i=0; i<strlen(src); i++) {
     sprintf(tmp, "%d ", (int)src[i]);
     strcat(dst, tmp);
   }
}
